#!/usr/bin/env python
import argparse
import tempfile

from Bio import PDB
from Bio.PDB.Polypeptide import aa1
from Bio.PDB.Polypeptide import aa3
from Bio.PDB.Polypeptide import one_to_three

from ssbio.protein.structure.utils.cleanpdb import CleanPDB
from ssbio.protein.structure.utils.structureio import StructureIO


class MutatePDB(PDB.Select):
    """Selection rules to mutate a PDB file

    These rules aim to:
    - Mutate a specified residue number to a new amino acid
    """

    keep_atom_list = ['N', 'C', 'O', 'CA']

    def __init__(self, mutation_list):
        """Initialize the parameters which indicate what mutations will occur

        Args:
            chain:
            residue_number:
            mutate_to:
        """
        self.mutation_list = [(i[0], int(i[1]), self._standard_resname(i[2])) for i in mutation_list]
        self.chains_and_residues = [(i[0], int(i[1])) for i in mutation_list]

    def _standard_resname(self, res):
        resname3 = res.upper()
        if resname3 not in list(aa3) and resname3 not in list(aa1):
            # TODO: mutation to selenocysteine (U;SEC) is not working
            raise ValueError("Unrecognised residue {}".format(res))
        if len(resname3) == 1:
            resname3 = one_to_three(resname3)

        return resname3

    def accept_residue(self, residue):
        hetfield, resseq, icode = residue.get_id()

        chain = residue.get_parent()
        chain_id = chain.get_id()

        if (chain_id,resseq) in self.chains_and_residues:
            prev_resname = residue.resname
            get_index = self.chains_and_residues.index((chain_id,resseq))
            residue.resname = self.mutation_list[get_index][2]
            print("Mutated {0}.{1}.{2} to {0}.{1}.{3}".format(chain_id, resseq, prev_resname, residue.resname))
        return True

    def accept_atom(self, atom):
        residue = atom.get_parent()
        hetfield, resseq, icode = residue.get_id()

        chain = residue.get_parent()
        chain_id = chain.get_id()

        if (chain_id,resseq) in self.chains_and_residues and atom.get_id() not in self.keep_atom_list:
            # print("Removing atom {}.{}.{}".format(chain_id, resseq, atom.get_id()))
            return False

        return True


def parse_mutation_input(instr):
    init_split = instr.split(',')
    second_split = [tuple(i.split('.')) for i in init_split]
    return second_split

if __name__ == '__main__':
    p = argparse.ArgumentParser(description='Mutates a PDB file')
    p.add_argument('infile', help='PDB file you want to mutate')
    p.add_argument('mutations', help='Mutations in the form of Chain1.ResNum1.Mutation1,Chain2.ResNum2.Mutation2. Example: A.4.TYR,B.4.TYR')
    p.add_argument('--outsuffix', '-o', default='_mutated', help='Suffix appended to PDB file')
    p.add_argument('--clean', '-c', action='store_true', help='Clean PDB and keep only chain with mutation')
    args = p.parse_args()

    mutations = parse_mutation_input(args.mutations)

    my_pdb = StructureIO(args.infile)
    if args.clean:
        my_cleaner = CleanPDB(keep_chains=[x[0] for x in mutations])
        my_clean_pdb = my_pdb.write_pdb(out_suffix='_clean', out_dir=tempfile.gettempdir(), custom_selection=my_cleaner)
        my_pdb = StructureIO(my_clean_pdb)

    my_mutation = MutatePDB(mutations)
    my_mutated_pdb = my_pdb.write_pdb(out_suffix=args.outsuffix, out_dir='mutated_pdbs', custom_selection=my_mutation)
    print('Mutated PDB at: {}'.format(my_mutated_pdb))



import warnings
from Bio import SeqIO
import pandas as pd

from ssbio.databases import uniprot

def standard_isoform_id(uni_iso):
    iso_num = uni_iso.split('-')[1]
    return iso_num.isdigit()

def gene_isos_to_uniprot(gene, isos, reviewed_uni_isos, unreviewed_uni_isos):
    mapping = {}
    reviewed_uni_isos = sorted(reviewed_uni_isos)
    unreviewed_uni_isos = sorted(unreviewed_uni_isos)
    combolist = unique_uniprot_isoforms(reviewed_uni_isos, unreviewed_uni_isos)

    for x in combolist:
        if not standard_isoform_id(x):
            raise ValueError('Strange isoform ID: {}'.format(x))

    # check if the number of reported isoforms is less than or equal to what's available
    if len(isos) <= len(combolist):
        for i,iso in enumerate(isos):
            mapping[iso] = combolist[i]
        return mapping
    # if there are more isoforms in the model than available in uniprot, just return None for that specific gene
    elif len(isos) > len(combolist):
        for i, iso in enumerate(isos):
            try:
                mapping[iso] = combolist[i]
            except:
                mapping[iso] = None
        return mapping

def unique_uniprot_isoforms(rev_isos, unrev_isos):
    """Filter out a list of uniprot isoforms"""
    final_list_of_unique_isoforms = []

    rev_isos = sorted(rev_isos)
    unrev_isos = sorted(unrev_isos)

    if len(unrev_isos) == 0:
        return rev_isos

    for unrev in unrev_isos:
        unique_unrev = True
        unrev_seq = uniprot.get_fasta(unrev)
        for rev in rev_isos:
            final_list_of_unique_isoforms.append(rev)
            if unrev_seq == uniprot.get_fasta(rev):
                unique_unrev = False
                break
        if unique_unrev:
            final_list_of_unique_isoforms.append(unrev)

    return final_list_of_unique_isoforms

def gene_to_uniprot(gene, isoforms, mapping_df, isoform_df, metadata_df):
    '''
    Input: a gene, its isoforms - 314, [1,2]
    Output: the mapped UniProt IDs to each of the isoforms as a dictionary {'314.1': 'O75106-1', '314.2': 'O75106-2'}

    Checks if there are UniProt ID mappings for a gene and its isoforms. A couple cases to consider:
    1. (Perfect) There are the same number of isoforms in UniProt as the entry query. The correct ones map to the correct numbers (ie .1 -> -1 and .2 -> -2)
    2. (Imperfect) The isoform IDs in UniProt are odd (-A and -B, or -3 is the main isoform)
        ## MAYBE NOT TODO: how to check this one? is there an isoform comment? probably have to manually read that. but there don't seem to be any!! yay!
        Output: {'314.1': 'O75106-A', '314.2': 'O75106-B'}
    3. (Missing) There are a lesser number UniProt isoforms for the gene.
        3a. Because the other isoform is in an unreviewed entry
            Output: {'314.1': 'O75106-1', '314.2': 'A0SDF98JH-1'}
        3b. Because UniProt doesn't have it
            Output: {'314.1': 'O75106-1', '314.2': None}
            # TODO: is this the best way to do it?
    4. (Missing) There is no UniProt entry
        Output: None
    5. (Other) There are multiple reviewed entries. These must be manually handled.

    ##TODO: please draw a diagram or something to explain what's going on here
    '''
    g_to_u_df = mapping_df[mapping_df.m_gene_entrez == gene]
    g_to_u_all = g_to_u_df.u_uniprot_acc.unique().tolist()
    g_to_u_reviewed = g_to_u_df[g_to_u_df.u_reviewed == True].u_uniprot_acc.unique().tolist()
    g_to_u_unreviewed = g_to_u_df[g_to_u_df.u_reviewed == False].u_uniprot_acc.unique().tolist()

    # if we cannot map to reviewed or unreviewed, we must give up
    if len(g_to_u_reviewed) == 0 and len(g_to_u_unreviewed) == 0:
#         print 'ah'
        warnings.warn('No reviewed or unreviewed UniProt entries for gene %s and its isoforms %s' % (gene, isoforms), UserWarning)
        return None

    # now check the number of isoforms
    if len(g_to_u_reviewed) == 1:
        g_to_u_reviewed_match = g_to_u_reviewed[0]

        # load the isoforms
        mapped_isoforms_df = isoform_df[isoform_df.u_uniprot_acc == g_to_u_reviewed_match]
        mapped_uni_isoforms = mapped_isoforms_df.u_isoform_id.tolist()
        mapping_dict = {}

        # check if the number of isoforms available is greater than or equal to the number of expected isoforms
        if len(mapped_uni_isoforms) >= len(isoforms):
            # for each of the recon 2 isoform ids...
            for isoform_id in isoforms:
                # create a hypothetical uniprot isoform
                ssb_uni_isoform = str(g_to_u_reviewed_match) + '-' + str(isoform_id)
                # check to see if it exists
                if ssb_uni_isoform in mapped_uni_isoforms:
                    mapping_dict[str(gene) + '.' + str(isoform_id)] = ssb_uni_isoform
                # if it doesn't, and there is only one isoform, that means there is an issue with the
                # metadata retrieval. Take the original uniprot ID as the mapping
                elif len(isoforms) == 1:
                    mapping_dict[str(gene) + '.' + str(isoform_id)] = g_to_u_reviewed_match + '-' + isoform_id
                else:
                    warnings.warn('TODO: unknown error', UserWarning)
                    return None
            return mapping_dict

        # if the number of isoforms available is less than the number of expected isoforms..
        # we first check the unreviewed uniprots
        elif len(g_to_u_unreviewed) >= 1:
            # get the sequence of the reviewed uniprot (from metadata DF just to make sure..)
            rev_uni_seqs = isoform_df[isoform_df.u_uniprot_acc == g_to_u_reviewed_match].u_seq.unique().tolist()
            # for the unreviewed entries, check if they are different sequences!
            mini_len_dict = {}
            # make sure to make a copy of the list since we are removing things..
            for unrev_uni in g_to_u_unreviewed[:]:
                unrev_uni_seq = metadata_df.loc[unrev_uni].u_seq
#                 print unrev_uni, unrev_uni_seq
                if unrev_uni_seq in rev_uni_seqs:
                    # remove the same sequences
#                     print 'REMOVE', unrev_uni
                    g_to_u_unreviewed.remove(unrev_uni)
                else:
                    # record the length for later use
#                     print 'KEEP', unrev_uni, len(unrev_uni_seq)
                    mini_len_dict[unrev_uni] = len(unrev_uni_seq)

            # now if our collection of unreviewed and reviewed entries is greater than the number of isoforms...
            if len(g_to_u_unreviewed) + len(mapped_uni_isoforms) >= len(isoforms):
                # match the reviewed isoforms to the genes like above..
                # and if the hypothetical uniprot isoform does not exist, match it to the longest unreviewed isoform

                # first sort the list of unreviewed isoforms..
                sorted_unrev_uni = [y[0] for y in sorted(mini_len_dict.items(), key=lambda x: x[1])]
                mapping_dict = {}
#                 print gene, isoforms, mapped_uni_isoforms,mini_len_dict,sorted_unrev_uni
                for isoform_id in isoforms:
                    # first map to reviewed entries
                    ssb_uni_isoform = g_to_u_reviewed_match + '-' + isoform_id
                    if ssb_uni_isoform in mapped_uni_isoforms:
                        mapping_dict[str(gene) + '.' + str(isoform_id)] = ssb_uni_isoform

                    # then map to longest unreviewed
                    else:
                        if len(sorted_unrev_uni) == 0:
                            break
                        # print sorted_unrev_uni
                        mapping_dict[str(gene) + '.' + str(isoform_id)] = sorted_unrev_uni[-1]
                        sorted_unrev_uni.pop(-1)
#                 print mapping_dict
                return mapping_dict

            # if our collection of unreviewed and reviewed entries is still less than the number of isoforms...
            # just map as many as we can
            else:
                for isoform_id in isoforms:
                    # create a hypothetical uniprot isoform
                    ssb_uni_isoform = g_to_u_reviewed_match + '-' + isoform_id
                    # check to see if it exists
                    if ssb_uni_isoform in mapped_uni_isoforms:
                        mapping_dict[str(gene) + '.' + str(isoform_id)] = ssb_uni_isoform
                    # if it doesn't, and there is only one isoform, that means there is an issue with the metadata retrieval. Take the original uniprot ID as the mapping
                    elif len(isoforms) == 1:
                        mapping_dict[str(gene) + '.' + str(isoform_id)] = g_to_u_reviewed_match
                    else:
                        mapping_dict[str(gene) + '.' + str(isoform_id)] = None
                return mapping_dict
        # if we just have a reviewed entry, and it has not enough isoforms
        # just map as many as we can
        else:
            for isoform_id in isoforms:
                # create a hypothetical uniprot isoform
                ssb_uni_isoform = g_to_u_reviewed_match + '-' + isoform_id
                # check to see if it exists
                if ssb_uni_isoform in mapped_uni_isoforms:
                    mapping_dict[str(gene) + '.' + str(isoform_id)] = ssb_uni_isoform
                # if it doesn't, and there is only one isoform, that means there is an issue with the metadata retrieval. Take the original uniprot ID as the mapping
                elif len(isoforms) == 1:
                    mapping_dict[str(gene) + '.' + str(isoform_id)] = g_to_u_reviewed_match
                else:
                    mapping_dict[str(gene) + '.' + str(isoform_id)] = None
            return mapping_dict
    elif len(g_to_u_reviewed) == 0 and len(g_to_u_unreviewed) >= len(isoforms):
        mapping_dict = {}
        mini_len_dict = {}
        # get the length of the unreviewed sequences
        for unrev_uni in g_to_u_unreviewed[:]:
            unrev_uni_seq = metadata_df.loc[unrev_uni].u_seq
            mini_len_dict[unrev_uni] = len(unrev_uni_seq)
            
        sorted_unrev_uni = [y[0] for y in sorted(mini_len_dict.items(), key=lambda x: x[1])]
        for isoform_id in isoforms:
            mapping_dict[str(gene) + '.' + str(isoform_id)] = sorted_unrev_uni[-1]
            sorted_unrev_uni.pop(-1)
        return mapping_dict
        
        

import glob
SEQ_SUNPRO_FILES = '/home/nathan/projects/homology_models/HSAPIENS/SUNPRO/refseq_sequences/'
available_refseq_files = glob.glob(SEQ_SUNPRO_FILES + '*.faa')
available_refseq = {}
for f in available_refseq_files:
    refseq_id = f.split('.faa')[0].split('/')[-1].split('.')[0]
    available_refseq[refseq_id] = f.split('/')[-1]

def gene_to_refseq(gene, isoforms, mapping_df):#, metadata_df):
    '''
    Input: a gene, its isoforms (314, [1,2]), mapping_df (the gene to refseq df with no isoforms), metadata_df (the refseq df with sequence and whatnot)
    Output: the mapped RefSeq IDs to each of the isoforms as a dictionary {'314.1': 'NP_033720', '314.2': 'NP_001149'}

    Checks if there are RefSeq ID mappings for a gene and its isoforms. A couple cases to consider:
    1. (Perfect) There are the same (or greater) number of isoforms in RefSeq as the entry query. The correct ones map to the correct transcript names (ie .1 -> -001 and .2 -> -002)
    2. (Imperfect) The isoform IDs in RefSeq are odd (-201 and -202, or -5 is the main isoform)
        ## NOTTODO: how to check this one? is there an isoform comment? seems like we have to do this manually...
        ## ANSWER: just map by length of sequence! https://www.biostars.org/p/19029/ - longest isoform is "canonical"
        Output: {'314.1': 'NP_033720', '314.2': 'NP_001149'}
    3. (Missing) There is only one RefSeq isoform for the mapped entry when there should be more.
        Output: None
    4. (Missing) There is no RefSeq entry
        Output: None
    5. (Other) There are multiple entries. These must be manually handled.
    '''

    isoforms = sorted(isoforms)
    g_to_many_df = mapping_df[mapping_df.bm_entrez == gene]
    g_to_r = g_to_many_df[pd.notnull(g_to_many_df.bm_refseq)].bm_refseq.unique().tolist()

    if len(g_to_r) == 0:
#         warnings.warn('WARNING: no mapped RefSeq IDs for gene %kegg and isoforms %kegg' % (gene, isoforms))
        return None

    # load the sequences of the isoforms, record their length. assign them to the isoforms based on length
    refseq_lengths = {}
    refseq_seqs = []
    for mapped_r in g_to_r:
        if mapped_r in available_refseq.keys():
            mapped_r_file = available_refseq[mapped_r]
            handle = open(SEQ_SUNPRO_FILES + mapped_r_file, "rU")
            for record in SeqIO.parse(handle, "fasta") :
                if record.seq not in refseq_seqs:
                    refseq_lengths[mapped_r] = len(record.seq)
                    refseq_seqs.append(record.seq)
            handle.close()
    sorted_refseqs = [y[0] for y in sorted(refseq_lengths.items(), key=lambda x: x[1])]
    if len(list(refseq_lengths.keys())) >= len(isoforms):
        mapping_dict = {}
        for iso in isoforms:
            mapping_dict[str(gene) + '.' + str(iso)] = sorted_refseqs.pop(-1)
        return mapping_dict
    
    # if the number of refseq ids is less than the number of isoforms
    # we should not remove the sequences if they are the same - they are just probably protein sequences with different UTRs or something
    elif len(list(refseq_lengths.keys())) < len(isoforms):
        
        # redo the above code but add even if the sequence is in there
        refseq_lengths = {}
        transcript_names = {}
        refseq_seqs = []
        for mapped_r in g_to_r:
            if mapped_r in available_refseq.keys():
                mapped_r_file = available_refseq[mapped_r]
                handle = open(SEQ_SUNPRO_FILES + mapped_r_file, "rU")
                for record in SeqIO.parse(handle, "fasta") :
                    refseq_lengths[mapped_r] = len(record.seq)
                    transcript_names[mapped_r] = g_to_many_df[g_to_many_df.bm_refseq == mapped_r].bm_name.unique()[0]
                    refseq_seqs.append(record.seq)
                handle.close()
#         sorted_refseqs = [y[0] for y in sorted(refseq_lengths.items(), key=lambda x: x[1])]
        # assign based on transcript name?
        sorted_refseqs = [y[0] for y in sorted(transcript_names.items(), key=lambda x: x[1], reverse=True)]
        
        mapping_dict = {}
        for iso in isoforms:
            if len(sorted_refseqs) == 0:
                mapping_dict[str(gene) + '.' + str(iso)] = None
            else:
                mapping_dict[str(gene) + '.' + str(iso)] = sorted_refseqs.pop(-1)
        return mapping_dict



import logging
import os
import os.path as op
import shutil
import time

import pandas as pd
from ssbio.protein.structure.utils.cleanpdb import CleanPDB
from ssbio.protein.structure.utils.structureio import StructureIO

import ssbio.utils
from ssbio.protein.structure.structprop import StructProp

log = logging.getLogger(__name__)


class ITASSERProp(StructProp):
    """Parse all available information for an I-TASSER modeling run.
    """

    # TODO: parse   1) predicted SS   2) predicted solvent acc    3) B-factor
    _main_files_to_copy = ['seq.dat', 'cscore']
    _coach_files_to_copy = ['Bsites.inf', 'EC.dat', 'GO_MF.dat', 'GO_BP.dat', 'GO_CC.dat']

    def __init__(self, ident, original_results_path, create_dfs=False,
                 coach_results_folder='model1/coach', model_to_use='model1'):
        """Initialize a class to collect I-TASSER modeling information and optionally copy results to a new directory.

        Args:
            original_results_path: Path to I-TASSER modeling folder
            create_dfs: If data frames should be created for COACH results
            coach_results_folder: Path to original COACH results
            model_to_use: Which I-TASSER model to use. Default is "model1"
        """
        if not op.exists(original_results_path):
            raise OSError('{}: folder does not exist'.format(original_results_path))

        StructProp.__init__(self, ident, is_experimental=False)

        self.results_path = original_results_path
        self.model_to_use = model_to_use
        self.create_dfs = create_dfs

        ### MODEL1.PDB THINGS
        old_model_path = op.join(original_results_path, '{}.pdb'.format(model_to_use))
        if not op.exists(old_model_path):
            log.debug('{}: No homology model available'.format(old_model_path))
            self.model_date = None
            self.model_file = None
        else:
            self.load_structure_path(old_model_path, 'pdb')
            # Save model creation date
            self.model_date = time.strftime('%Y-%m-%d', time.gmtime(os.path.getmtime(old_model_path)))
            self.model_file = op.basename(self.structure_path)

        ### MODELING RESULTS
        # Parse init.dat
        init_dat_path = op.join(original_results_path, 'init.dat')
        if op.exists(init_dat_path):
            self.update(self.parse_init_dat(init_dat_path))

        # Parse seq.dat
        seq_dat_path = op.join(original_results_path, 'seq.dat')
        if op.exists(seq_dat_path):
            pass
            # TODO: parse seq.dat and store in modeling_results

        # Parse cscore
        cscore_path = op.join(original_results_path, 'cscore')
        if op.exists(cscore_path):
            self.update(self.parse_cscore(cscore_path))

        # Parse COACH results if available
        coach_folder = op.join(original_results_path, coach_results_folder)
        if op.isdir(coach_folder):

            # Parse Bsites.inf
            bsites_inf_path = op.join(coach_folder, 'Bsites.inf')
            if op.exists(bsites_inf_path):
                parsed_bsites = self.parse_bsites_inf(infile=bsites_inf_path, create_df=create_dfs)
                if parsed_bsites:
                    tmp = {'top_bsite_'+k: v for k, v in parsed_bsites[0].items()}
                    self.update(tmp)

            # Parse EC.dat
            ec_dat_path = op.join(coach_folder, 'EC.dat')
            if op.exists(ec_dat_path):
                parsed_ec = self.parse_ec(infile=ec_dat_path, create_df=create_dfs)
                if parsed_ec:
                    tmp = {'top_ec_' + k: v for k, v in parsed_ec[0].items()}
                    self.update(tmp)

            # Parse GO_MF.dat
            go_mf_dat_path = op.join(coach_folder, 'GO_MF.dat')
            if op.exists(go_mf_dat_path):
                parsed_go_mf = self.parse_go(infile=go_mf_dat_path, create_df=create_dfs)
                if parsed_go_mf:
                    tmp = {'top_go_mf_' + k: v for k, v in parsed_go_mf[0].items()}
                    self.update(tmp)

            # Parse GO_BP.dat
            go_bp_dat_path = op.join(coach_folder, 'GO_BP.dat')
            if op.exists(go_bp_dat_path):
                parsed_go_bp = self.parse_go(infile=go_bp_dat_path, create_df=create_dfs)
                if parsed_go_bp:
                    tmp = {'top_go_bp_' + k: v for k, v in parsed_go_bp[0].items()}
                    self.update(tmp)

            # Parse GO_CC.dat
            go_cc_dat_path = op.join(coach_folder, 'GO_CC.dat')
            if op.exists(go_cc_dat_path):
                parsed_go_cc = self.parse_go(infile=go_cc_dat_path, create_df=create_dfs)
                if parsed_go_cc:
                    tmp = {'top_go_cc_' + k: v for k, v in parsed_go_cc[0].items()}
                    self.update(tmp)

    def copy_results(self, copy_to_dir, rename_model_to=None, force_rerun=False):
        """Copy the raw information from I-TASSER modeling to a new folder.

        Copies files in the variables _main_files_to_copy and _coach_files_to_copy.

        Args:
            copy_to_dir: Directory to copy the minimal set of results per sequence.
            rename_model_to: New file name (without extension)
            force_rerun: If existing models and results should be overwritten.

        """
        # Save path to the structure and copy it if specified
        if not rename_model_to:
            rename_model_to = self.model_to_use

        new_model_path = op.join(copy_to_dir, '{}.pdb'.format(rename_model_to))

        if self.structure_path:
            if ssbio.utils.force_rerun(flag=force_rerun, outfile=new_model_path):
                # Clean and save it
                custom_clean = CleanPDB()
                my_pdb = StructureIO(self.structure_path)
                new_model_path = my_pdb.write_pdb(custom_selection=custom_clean,
                                                  custom_name=rename_model_to,
                                                  out_dir=copy_to_dir,
                                                  force_rerun=force_rerun)
            # Update the structure_path to be the copied, clean file
            self.load_structure_path(new_model_path, 'pdb')
            self.model_file = op.basename(new_model_path)

            # Other modeling results - store in a new folder
            dest_itasser_dir = op.join(copy_to_dir, '{}_itasser'.format(rename_model_to))
            if not op.exists(dest_itasser_dir):
                os.mkdir(dest_itasser_dir)

            for f in self._main_files_to_copy:
                old_file_path = op.join(self.results_path, f)
                new_file_path = op.join(dest_itasser_dir, f)
                if op.exists(old_file_path):
                    if ssbio.utils.force_rerun(flag=force_rerun, outfile=new_file_path):
                        shutil.copy2(old_file_path, new_file_path)
                        log.debug('{}: copied from {}'.format(new_file_path, old_file_path))
                    else:
                        log.debug('{}: file already exists'.format(new_file_path))

            for ff in self._coach_files_to_copy:
                old_file_path = op.join(self.results_path, 'model1/coach/', ff)
                new_file_path = op.join(dest_itasser_dir, ff)
                if op.exists(old_file_path):
                    if ssbio.utils.force_rerun(flag=force_rerun, outfile=new_file_path):
                        shutil.copy2(old_file_path, new_file_path)
                        log.debug('{}: copied from {}'.format(new_file_path, old_file_path))
                    else:
                        log.debug('{}: file already exists'.format(new_file_path))

    def save_dataframes(self, outdir):
        """Save all attributes that start with "df_" into a specified directory.

        Args:
            outdir: Path to output directory.

        Returns:

        """
        # TODO: test this method
        if not self.create_dfs:
            raise NameError('Nothing to save! Create dataframes by specifying create_dfs=True.')
        else:
            super(StructProp, self).save_dataframes(outdir)

    def parse_init_dat(self, infile):
        """Parse the main init.dat file which contains the modeling results

        The first line of the file init.dat contains stuff like:
        "120 easy  40   8"
        The other lines look like this:
        "     161   11.051   1  1guqA MUSTER"
        and getting the first 10 gives you the top 10 templates used in modeling

        Args:
            infile: Path to init.dat
            copy_to_folder: Optional folder to copy init.dat to

        Returns:
            dict: Dictionary of parsed information

        """
        init_dict = {}

        log.debug('{}: Reading file...'.format(infile))
        with open(infile, 'r') as f:
            # Get first 2 lines of file
            head = [next(f).strip() for x in range(2)]

        summary = head[0].split()
        difficulty = summary[1]

        top_template_info = head[1].split()
        top_template_pdbchain = top_template_info[3]
        top_template_pdb = top_template_pdbchain[:4]
        top_template_chain = top_template_pdbchain[4:]

        init_dict['difficulty'] = difficulty
        init_dict['top_template_pdb'] = top_template_pdb
        init_dict['top_template_chain'] = top_template_chain

        return init_dict

    def parse_seq_dat(self, infile):
        """Parse the secondary structure predictions in seq.dat

        Args:
            infile: Path to seq.dat

        Returns:
            dict: Dictionary of residue number to secondary structure

        """
        pass

    def parse_cscore(self, infile):
        """Parse the cscore file to return a dictionary of scores.

        Args:
            infile: Path to cscore

        Returns:
            dict: Dictionary of scores

        """
        cscore_dict = {}
        with open(infile, 'r') as f:
            for ll in f.readlines():
                # Look for the first line that starts with model1
                if ll.lower().startswith('model1'):
                    l = ll.split()

                    cscore = l[1]
                    tmscore_full = l[2].split('+-')
                    tmscore = tmscore_full[0]
                    tmscore_err = tmscore_full[1]
                    rmsd_full = l[3].split('+-')
                    rmsd = rmsd_full[0]
                    rmsd_err = rmsd_full[1]

                    cscore_dict['c_score'] = float(cscore)
                    cscore_dict['tm_score'] = float(tmscore)
                    cscore_dict['tm_score_err'] = float(tmscore_err)
                    cscore_dict['rmsd'] = float(rmsd)
                    cscore_dict['rmsd_err'] = float(rmsd_err)

        return cscore_dict

    def parse_bsites_inf(self, infile, create_df=False):
        """Parse the Bsites.inf output file of COACH and returns a dataframe

        Bsites.inf contains the summary of COACH clustering results after all
            other prediction algorithms have finished
        For each site (cluster), there are three lines:
        Line 1: site number, c-score of coach prediction, cluster size
        Line 2: algorithm, PDB ID, ligand ID, center of binding site (cartesian coordinates),
            c-score of the algorithm's prediction, binding residues from single template
        Line 3: Statistics of ligands in the cluster

        Args:
            infile (str): Path to Bsites.inf
            create_df (bool): If a Pandas DataFrame should be saved to the object under the attribute "df_bsites"

        Returns:
            list: Ranked list of dictionaries, keys defined below
                site_num: cluster which is the consensus binding site
                c_score: confidence score of the cluster prediction
                cluster_size: number of predictions within this cluster
                algorithm: main? algorithm used to make the prediction
                pdb_template_id: PDB ID of the template used to make the prediction
                pdb_template_chain: chain of the PDB which has the ligand
                pdb_ligand: predicted ligand to bind
                binding_location_coords: centroid of the predicted ligand position in the homology model
                c_score_method: confidence score for the main algorithm
                binding_residues: predicted residues to bind the ligand
                ligand_cluster_counts: number of predictions per ligand

        """

        pre_bsites_inf_df = []

        with open(infile) as pp:
            lines = list(filter(None, (line.rstrip() for line in pp)))

        for i in range(len(lines) // 3):
            bsites_site_dict = {}

            line1 = lines[i * 3].split('\t')
            line2 = lines[i * 3 + 1].split('\t')
            line3 = lines[i * 3 + 2]

            bsites_site_dict['site_num'] = line1[0]
            bsites_site_dict['c_score'] = float(line1[1])
            bsites_site_dict['cluster_size'] = line1[2]

            bsites_site_dict['algorithm'] = line2[0]
            bsites_site_dict['pdb_template_id'] = line2[1][:4]
            bsites_site_dict['pdb_template_chain'] = line2[1][4]
            bsites_site_dict['pdb_ligand'] = line2[2]
            bsites_site_dict['binding_location_coords'] = tuple(float(x) for x in line2[3].split())

            # TODO: what's the difference between this c-score and the cluster's c-score?
            # how is the cluster's c-score computed? it's not the average c-score of all methods
            # also why are some COFACTOR c-scores >1?
            # 160411 - seems like the COFACTOR "BS-score" is being reported here, not its c-score...
            tmp_split = line2[4].split(' :')
            bsites_site_dict['c_score_method'] = tmp_split[0]
            bsites_site_dict['binding_residues'] = tmp_split[1]

            bsites_site_dict['ligand_cluster_counts'] = line3

            pre_bsites_inf_df.append(bsites_site_dict)

        if create_df:
            cols = ['site_num', 'c_score', 'cluster_size', 'algorithm',
                    'pdb_template_id', 'pdb_template_chain', 'pdb_ligand',
                    'binding_location_coords', 'c_score_method', 'binding_residues',
                    'ligand_cluster_counts']
            bsites_inf_df = pd.DataFrame.from_records(pre_bsites_inf_df, columns=cols).drop_duplicates().reset_index(drop=True)
            bsites_inf_df['c_score'] = pd.to_numeric(bsites_inf_df.c_score, errors='coerce')
            bsites_inf_df['cluster_size'] = pd.to_numeric(bsites_inf_df.cluster_size, errors='coerce')

            self.df_bsites = bsites_inf_df
            log.debug('Created binding site prediction DataFrame in attribute "df_bsites')

        return pre_bsites_inf_df


    def parse_ec(self, infile, create_df=False):
        """Parse the EC.dat output file of COACH and returns a dataframe

        EC.dat contains the predicted EC number and active residues.
            The columns are: PDB_ID, TM-score, RMSD, Sequence identity,
            Coverage, Confidence score, EC number, and Active site residues

        Args:
            infile (str): Path to EC.dat

        Returns:
             list: Ranked list of dictionaries, keys defined below
                pdb_template_id: PDB ID of the template used to make the prediction
                pdb_template_chain: chain of the PDB which has the ligand
                tm_score: TM-score of the template to the model (similarity score)
                rmsd: RMSD of the template to the model (also a measure of similarity)
                seq_ident: percent sequence identity
                seq_coverage: percent sequence coverage
                c_score: confidence score of the EC prediction
                ec_number: predicted EC number
                binding_residues: predicted residues to bind the ligand

        """

        ec_df = pd.read_table(infile, delim_whitespace=True,
                              names=['pdb_template', 'tm_score', 'rmsd', 'seq_ident', 'seq_coverage',
                                     'c_score', 'ec_number', 'binding_residues'])
        ec_df['pdb_template_id'] = ec_df['pdb_template'].apply(lambda x: x[:4])
        ec_df['pdb_template_chain'] = ec_df['pdb_template'].apply(lambda x: x[4])

        ec_df = ec_df[['pdb_template_id', 'pdb_template_chain', 'tm_score', 'rmsd',
                       'seq_ident', 'seq_coverage', 'c_score', 'ec_number', 'binding_residues']]
        ec_df['c_score'] = pd.to_numeric(ec_df.c_score, errors='coerce')

        if create_df:
            self.df_ec = ec_df
            log.debug('Created EC prediction DataFrame in attribute "df_ec')

        return list(ec_df.to_dict(orient='index').values())


    def parse_go(self, infile, create_df=False):
        """Parse a GO output file from COACH and returns a dataframe

        The columns in all files are: GO terms, Confidence score, Name of GO terms.
            GO_MF.dat - GO terms in 'molecular function'
            GO_BP.dat - GO terms in 'biological process'
            GO_CC.dat - GO terms in 'cellular component'

        Args:
            infile (str): Path to any GO file

        Returns:
            Pandas DataFrame: Organized dataframe of results, columns defined below
                go_id: GO term ID
                go_term: GO term text
                c_score: confidence score of the GO prediction

        """
        go_list = []

        with open(infile) as go_file:
            for line in go_file.readlines():
                go_dict = {}

                go_split = line.split()
                go_dict['go_id'] = go_split[0]
                go_dict['c_score'] = go_split[1]
                go_dict['go_term'] = ' '.join(go_split[2:])

                go_list.append(go_dict)

        if create_df:
            cols = ['go_id', 'go_term', 'c_score']
            go_df = pd.DataFrame.from_records(go_list, columns=cols).drop_duplicates().reset_index(drop=True)
            go_df['c_score'] = pd.to_numeric(go_df.c_score, errors='coerce')

            if hasattr(self, 'df_go'):
                self.df_go.append(go_df)
            else:
                self.df_go = go_df
            log.debug('Created GO prediction DataFrame in attribute "df_go"')

        return go_list




import unittest
import os.path as op
import ssbio.utils

class TestUtils(unittest.TestCase):
    """Unit tests for utils
    """

    def test_split_folder_and_path(self):
        test_path = op.join('this','is','a','hypothetical','path','to_a_file.myfile')
        result = (op.join('this','is','a','hypothetical','path'), 'to_a_file', '.myfile')

        self.assertEqual(result, ssbio.utils.split_folder_and_path(test_path))

    def test_force_rerun(self):
        # TODO: need to use op.join for the files, for tests to work properly on windows
        self.assertTrue(ssbio.utils.force_rerun(flag=True, outfile=op.join('not','existing','file.txt')))

        self.assertTrue(ssbio.utils.force_rerun(flag=False, outfile=op.join('not','existing','file.txt')))

        self.assertTrue(ssbio.utils.force_rerun(flag=True, outfile=op.join('test_files','Ec_core_flux1.xml')))

        self.assertFalse(ssbio.utils.force_rerun(flag=False, outfile=op.join('test_files','Ec_core_flux1.xml')))

    def test_outfile_name_maker(self):
        test_out = ssbio.utils.outfile_maker(inname='P00001.fasta')
        correct_out = 'P00001.out'
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname='P00001.fasta', append_to_name='_new')
        correct_out = 'P00001_new.out'
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname='P00001.fasta', outext='.mao')
        correct_out = 'P00001.mao'
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname='P00001.fasta', outext='.mao', append_to_name='_new')
        correct_out = 'P00001_new.mao'
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname='P00001.fasta', outext='.new', outname='P00001_aligned')
        correct_out = 'P00001_aligned.new'
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname='P00001.fasta', outname='P00001_aligned')
        correct_out = 'P00001_aligned.out'
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname='P00001.fasta', outname='P00001_aligned', append_to_name='_new')
        correct_out = 'P00001_aligned_new.out'
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname='P00001.fasta', outname='P00001_aligned', outdir=op.join('my','dir'))
        correct_out = op.join('my','dir','P00001_aligned.out')
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname='P00001.fasta', outname='P00001_aligned', outdir=op.join('my','dir'), outext='.test')
        correct_out = op.join('my','dir','P00001_aligned.test')
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname=op.join('test','other','dir','P00001.fasta'), append_to_name='_new')
        correct_out = op.join('test','other','dir','P00001_new.out')
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname=op.join('test','other','dir','P00001.fasta'), outname='P00001_aligned')
        correct_out = op.join('test','other','dir','P00001_aligned.out')
        self.assertEqual(test_out, correct_out)

        test_out = ssbio.utils.outfile_maker(inname=op.join('test','other','dir','P00001.fasta'), outname='P00001_aligned', outdir=op.join('my','dir'))
        correct_out = op.join('my','dir','P00001_aligned.out')
        self.assertEqual(test_out, correct_out)


from cobra.core import Gene
from ssbio.core.protein import Protein
import os.path as op
import ssbio.utils
import cobra.io.json
from json_tricks.np import dumps, loads
import logging
log = logging.getLogger(__name__)

class GenePro(Gene):
    """Extends the COBRAPy Gene object to add GEM-PRO annotations in the form of a Protein attribute
    """

    def __init__(self, id, name='', functional=True, root_dir=None, pdb_file_type='cif'):
        Gene.__init__(self, id=id, name=name, functional=functional)
        self.protein = Protein(ident=id, root_dir=root_dir, pdb_file_type=pdb_file_type)

        self.root_dir = root_dir

    @property
    def root_dir(self):
        """Directory where Gene folder is located"""
        return self._root_dir

    @root_dir.setter
    def root_dir(self, path):
        self._root_dir = path

        if path:
            if not op.exists(path):
                raise ValueError('{}: folder does not exist'.format(path))

            ssbio.utils.make_dir(self.gene_dir)

            # Additionally set the protein folder path
            self.protein.root_dir = self.gene_dir

    @property
    def gene_dir(self):
        """Gene folder"""
        if self.root_dir:
            return op.join(self.root_dir, self.id)
        else:
            log.warning('Root directory not set')
            return None

    def reset_protein(self):
        self.protein = Protein(self.id)

    def __json_encode__(self):
        reqd_attribs = cobra.io.json._REQUIRED_GENE_ATTRIBUTES
        optn_attribs = cobra.io.json._OPTIONAL_GENE_ATTRIBUTES
        optn_attribs.update({'root_dir': None})
        new_gene = {key: str(getattr(self, key))
                    for key in reqd_attribs}
        cobra.io.json._update_optional(self, new_gene, optn_attribs)

        # Protein
        new_gene['protein'] = self.protein

        return new_gene

    def __json_decode__(self, **attrs):
        Gene.__init__(self, id=attrs['id'])
        self.protein = attrs['protein']
        for k, v in cobra.io.json._OPTIONAL_GENE_ATTRIBUTES.items():
            if k not in attrs:
                setattr(self, k, v)
        for k, v in attrs.items():
            if k != 'protein' or k != 'id':
                setattr(self, k, v)



import logging
from collections import defaultdict
from copy import deepcopy
from os import path as op
import requests
from Bio import SeqIO
from cobra.core import DictList
from slugify import Slugify
import ssbio.protein.sequence.properties.residues
import ssbio.protein.sequence.utils
import ssbio.protein.sequence.utils.fasta
import ssbio.databases.pdb
import ssbio.utils
from ssbio.core.object import Object

custom_slugify = Slugify(safe_chars='-_.')
log = logging.getLogger(__name__)


class SeqProp(Object):
    def __init__(self, ident, sequence_path=None, metadata_path=None, seq=None, description="<unknown description>",
                 write_fasta_file=False, outfile=None, force_rewrite=False):
        """Store basic protein sequence properties.

        Takes as input a sequence FASTA file, sequence string, Biopython Seq or SeqRecord. You can also provide the
        path to the metadata file if available (see UniProtProp and KEGGProp classes for parsers). If provided as a
        string/Seq/SeqRecord object, you can also write them out as FASTA files.

        Args:
            ident (str): Identifier of the sequence
            sequence_path (str): Path to FASTA file of the sequence
            metadata_path (str): Path to metadata file of the sequence
            seq (str, Seq, SeqRecord): Sequence string, Biopython Seq or SeqRecord object
            description (str): Optional description of the sequence
            write_fasta_file (bool): If a FASTA file of the sequence should be written
            outname:
            outdir:
            force_rewrite:
        """

        Object.__init__(self, id=ident, description=description)

        # Database identifiers
        self.bigg = None
        self.kegg = None
        self.refseq = None
        self.uniprot = None
        self.gene_name = None
        self.pdbs = None

        # Sequence information
        self._seq_record = None
        if seq:
            # Load sequence if not provided as a file
            self.seq_record = ssbio.protein.sequence.utils.cast_to_seq_record(obj=seq, id=ident, description=description)

        # File information
        self._sequence_dir = None
        self.sequence_file = None
        if sequence_path:
            self.load_sequence_path(sequence_path)
        self._metadata_dir = None
        self.metadata_file = None
        if metadata_path:
            self.load_metadata_path(metadata_path)
        if write_fasta_file:
            if not outfile:
                raise ValueError('Output path must be specified if you want to write a FASTA file.')
            self.write_fasta_file(outfile=outfile, force_rerun=force_rewrite)

        # Store AlignIO objects of this sequence to others in alignments
        self.sequence_alignments = DictList()
        self.structure_alignments = DictList()

        # Copy SeqRecord annotations and letter annotations for JSON saving capabilities
        self.annotations = {}
        self.letter_annotations = {}

    @property
    def sequence_dir(self):
        return self._sequence_dir

    @sequence_dir.setter
    def sequence_dir(self, path):
        if not op.exists(path):
            raise ValueError('{}: folder does not exist'.format(path))

        self._sequence_dir = path

    @property
    def sequence_path(self):
        if self.sequence_dir and self.sequence_file:
            path = op.join(self.sequence_dir, self.sequence_file)
            if not op.exists(path):
                raise ValueError('{}: file does not exist'.format(path))
            return path
        else:
            if not self.sequence_dir:
                log.debug('{}: sequence directory not set'.format(self.id))
            if not self.sequence_file:
                log.debug('{}: sequence file not available'.format(self.id))
            return None

    @property
    def seq_record(self):
        # The SeqRecord object is not kept in memory unless explicitly set, so when saving as a json we only save a
        # pointer to the file
        if not self.sequence_path:
            return self._seq_record
            # log.error('{}: sequence file not available'.format(self.id))
            # return None


        return SeqIO.read(open(self.sequence_path), 'fasta')

    @seq_record.setter
    def seq_record(self, sr):
        # Copy SeqRecord annotations and letter annotations for JSON saving capabilities
        if sr:
            self.annotations = sr.annotations
            self.letter_annotations = sr.letter_annotations
        self._seq_record = sr

    @property
    def metadata_dir(self):
        return self._metadata_dir

    @metadata_dir.setter
    def metadata_dir(self, path):
        if not op.exists(path):
            raise ValueError('{}: folder does not exist'.format(path))

        self._metadata_dir = path

    @property
    def metadata_path(self):
        if self.metadata_dir and self.metadata_file:
            path = op.join(self.metadata_dir, self.metadata_file)
            if not op.exists(path):
                raise ValueError('{}: file does not exist'.format(path))
            return path
        else:
            if not self.metadata_dir:
                log.debug('{}: metadata directory not set'.format(self.id))
            if not self.metadata_file:
                log.debug('{}: metadata file not available'.format(self.id))
            return None

    @property
    def seq_str(self):
        """Get the sequence formatted as a string"""
        if not self.seq_record:
            return None
        return ssbio.protein.sequence.utils.cast_to_str(self.seq_record)

    @property
    def seq_len(self):
        """Get the sequence length"""
        if not self.seq_record:
            return 0
        return len(self.seq_record.seq)

    @property
    def num_pdbs(self):
        """Report the number of PDB IDs mapped"""
        if not self.pdbs:
            return 0
        else:
            return len(self.pdbs)

    def load_sequence_path(self, sequence_path):
        """Load a sequence file and provide pointers to its location

        Args:
            sequence_path: Path to sequence file

        """
        self.sequence_dir = op.dirname(sequence_path)
        self.sequence_file = op.basename(sequence_path)

    def load_metadata_path(self, metadata_path):
        """Load a metadata file and provide pointers to its location

        Args:
            metadata_path: Path to metadata file

        """
        self.metadata_dir = op.dirname(metadata_path)
        self.metadata_file = op.basename(metadata_path)

    def write_fasta_file(self, outfile, force_rerun=False):
        """Write a FASTA file for the protein sequence.

        Args:
            outfile (str): Path to new FASTA file to be written to
            force_rerun (bool): If an existing file should be overwritten

        """
        outdir, outname, outext = ssbio.utils.split_folder_and_path(outfile)
        sequence_path = ssbio.protein.sequence.utils.fasta.write_fasta_file(self.seq_record,
                                                                            outname=outname,
                                                                            outdir=outdir,
                                                                            outext=outext,
                                                                            force_rerun=force_rerun)
        # Unset the SeqRecord as it will now be dynamically loaded from the file
        self.seq_record = None
        self.load_sequence_path(sequence_path)

    def equal_to(self, seq_prop):
        """Test if the sequence is equal to another SeqProp object's sequence

        Args:
            seq_prop: SeqProp object

        Returns:
            bool: If the sequences are the same

        """
        if not self.seq_str:
            return False

        if not seq_prop:
            return False

        if not seq_prop.seq_str:
            return False

        return self.seq_str == seq_prop.seq_str

    def equal_to_fasta(self, seq_file):
        """Test if this sequence is equal to another sequence file.

        Args:
            seq_file: Path to another sequence file

        Returns:
            bool: If the sequences are the same

        """
        if not self.sequence_path:
            log.error('Sequence file not available')
            return False
        return ssbio.protein.sequence.utils.fasta.fasta_files_equal(self.sequence_path, seq_file)

    def load_letter_annotations(self, annotation_name, letter_annotation):
        """Add per-residue annotations to the seq_record.letter_annotations attribute

        Args:
            annotation_name: Name of the key that will be used to access this annotation
            letter_annotation (str, list, tuple): Sequence that contains information on the sequence, with the length
                being equal to the protein sequence length

        """
        self.letter_annotations[annotation_name] = letter_annotation
        log.debug('{}: loaded letter_annotations and saved as "{}"'.format(self.id, annotation_name))

    def get_biopython_pepstats(self):
        """Run Biopython's built in ProteinAnalysis module.

        Stores statistics in the seq_record.annotations attribute.
        """
        try:
            pepstats = ssbio.protein.sequence.properties.residues.biopython_protein_analysis(self.seq_str)
        except KeyError as e:
            log.error('{}: unable to run ProteinAnalysis module, unknown amino acid {}'.format(self.id, e))
            return
        self.annotations.update(pepstats)

    def get_emboss_pepstats(self):
        """Run the EMBOSS pepstats program on the protein sequence.

        Stores statistics in the seq_record.annotations attribute.
        Saves a .pepstats file of the results where the sequence file is located.
        """
        outfile = ssbio.protein.sequence.properties.residues.emboss_pepstats_on_fasta(infile=self.sequence_path)
        pepstats = ssbio.protein.sequence.properties.residues.emboss_pepstats_parser(outfile)
        self.annotations.update(pepstats)

    def blast_pdb(self, seq_ident_cutoff=0, evalue=0.0001, display_link=False,
                  outdir=None, force_rerun=False):
        """BLAST this sequence to the PDB"""
        if not outdir:
            outdir = self.sequence_dir
            if not outdir:
                raise ValueError('Output directory must be specified')

        if not self.seq_str:
            log.error('{}: no sequence loaded'.format(self.id))
            return None

        try:
            blast_results = ssbio.databases.pdb.blast_pdb(self.seq_str,
                                                          outfile='{}_blast_pdb.xml'.format(custom_slugify(self.id)),
                                                          outdir=outdir,
                                                          force_rerun=force_rerun,
                                                          evalue=evalue,
                                                          seq_ident_cutoff=seq_ident_cutoff,
                                                          link=display_link)
        except requests.ConnectionError as e:
            print(e)
            log.error('{}: BLAST request timed out'.format(seq_prop.id))
            return None

        return blast_results

    def summary(self, as_df):
        """Summarize this sequence.

        Returns:
            dict:

        """
        pass

    def sequence_mutation_summary(self):
        """Summarize all mutations found in the sequence_alignments attribute.

        Returns 2 dictionaries, single_counter and fingerprint_counter.

        single_counter:
            Dictionary of {point mutation: list of genes/strains}
            Example: {('A', 24, 'V'): ['Strain1', 'Strain2', 'Strain4'],
                      ('R', 33, 'T'): ['Strain2']}
            Here, we report which genes/strains have the single point mutation.

        fingerprint_counter:
            Dictionary of {mutation group: list of genes/strains}
            Example: {(('A', 24, 'V'), ('R', 33, 'T')): ['Strain2'],
                      (('A', 24, 'V')): ['Strain1', 'Strain4']}
            Here, we report which genes/strains have the specific combinations (or "fingerprints") of point mutations

        Returns:
            dict, dict: single_counter, fingerprint_counter

        """
        if len(self.sequence_alignments) == 0:
            log.error('{}: no sequence alignments'.format(self.id))
            return {}, {}

        fingerprint_counter = defaultdict(list)
        single_counter = defaultdict(list)

        for alignment in self.sequence_alignments:
            other_sequence = alignment.annotations['b_seq']
            mutations = alignment.annotations['mutations']

            if mutations:
                # Turn this list of mutations into a tuple so it can be a dictionary key
                mutations = tuple(tuple(x) for x in mutations)
                fingerprint_counter[mutations].append(other_sequence)

                for m in mutations:
                    single_counter[m].append(other_sequence)

        return dict(single_counter), dict(fingerprint_counter)

    def get_dict(self, only_keys=None, exclude_attributes=None, df_format=False):
        """Get a copy of all attributes as a dictionary, including object properties

        Args:
            only_keys:
            exclude_attributes:
            df_format:

        Returns:

        """
        my_dict = Object.get_dict(self, only_keys=only_keys, exclude_attributes=exclude_attributes, df_format=df_format)

        additional_keys = ['seq_str', 'seq_len', 'sequence_file', 'metadata_file', 'num_pdbs']
        for k in additional_keys:
            my_dict[k] = deepcopy(getattr(self, k))

        # Choose attributes to return, return everything in the object if a list is not specified
        if not only_keys:
            keys = list(my_dict.keys())
        else:
            keys = ssbio.utils.force_list(only_keys)

        # Remove keys you don't want returned
        if exclude_attributes:
            exclude_attributes = ssbio.utils.force_list(exclude_attributes)
            for x in exclude_attributes:
                if x in keys:
                    keys.remove(x)

        # Copy attributes into a new dictionary
        for k, v in my_dict.items():
            if k in keys:
                if df_format:
                    if v and not isinstance(v, str) and not isinstance(v, int) and not isinstance(v, float) and not isinstance(v, bool):
                        try:
                            my_dict[k] = ssbio.utils.force_string(v)
                        except TypeError:
                            log.warning('{}: excluding attribute from dict, cannot transform into string'.format(k))
                    else:
                        my_dict[k] = v
                else:
                    my_dict[k] = v
        return my_dict

    def __json_decode__(self, **attrs):
        for k, v in attrs.items():
            if k == 'sequence_alignments' or k == 'structure_alignments':
                setattr(self, k, DictList(v))
            else:
                setattr(self, k, v)


import logging
import os.path as op

import numpy as np
import ssbio.protein.structure.properties
from cobra.core import DictList
from ssbio.protein.structure.utils.structureio import StructureIO

import ssbio.protein.sequence.utils.alignment
import ssbio.utils
from ssbio.core.object import Object
from ssbio.protein.structure.chainprop import ChainProp

if ssbio.utils.is_ipynb():
    import nglview as nv
import seaborn as sns
import ssbio.protein.structure.utils.cleanpdb
log = logging.getLogger(__name__)


class StructProp(Object):
    """Class for protein structural properties"""

    def __init__(self, ident, description=None, chains=None, mapped_chains=None, structure_path=None, file_type=None,
                 representative_chain=None, is_experimental=False):
        Object.__init__(self, id=ident, description=description)

        self.is_experimental = is_experimental

        self.reference_seq_top_coverage = None

        # Chain information
        # chains is a DictList of ChainProp objects
        self.chains = DictList()
        if chains:
            self.add_chain_ids(chains)
        # representative_chain is a pointer to the chain in self.chains that matches reference_seq
        self.representative_chain = None
        if representative_chain:
            self.representative_chain = self.chains.get_by_id(representative_chain)
        # mapped_chains is an ordered list of mapped chain IDs which would come from BLAST or the best_structures API
        self.mapped_chains = ssbio.utils.force_list(mapped_chains)

        # File information
        self.file_type = file_type
        self._structure_dir = None
        self.structure_file = None
        if structure_path:
            self.load_structure_path(structure_path, file_type)

    @property
    def structure_dir(self):
        return self._structure_dir

    @structure_dir.setter
    def structure_dir(self, path):
        if not op.exists(path):
            raise ValueError('{}: folder does not exist'.format(path))

        self._structure_dir = path

    @property
    def structure_path(self):
        if self.structure_dir and self.structure_file:
            path = op.join(self.structure_dir, self.structure_file)
            if not op.exists(path):
                raise ValueError('{}: file does not exist'.format(path))
            return path
        else:
            if not self.structure_dir:
                log.debug('{}: structure directory not set'.format(self.id))
            if not self.structure_file:
                log.debug('{}: structure file not available'.format(self.id))
            return None

    def load_structure_path(self, structure_path, file_type):
        """Load a structure file and provide pointers to its location

        Args:
            structure_path: Path to structure file
            file_type: Type of structure file
        """
        self.file_type = file_type
        self.structure_dir = op.dirname(structure_path)
        self.structure_file = op.basename(structure_path)

    def parse_structure(self):
        """Read the 3D coordinates of a structure file and return it as a Biopython Structure object

        Also create ChainProp objects in the chains attribute

        Returns:
            Structure: Biopython structure object

        """
        if not self.structure_path:
            log.error('{}: no structure file, unable to parse'.format(self.id))
            return None
        else:
            # Add Biopython structure object
            structure = StructureIO(self.structure_path)

            # Add all chains to self.chains as ChainProp objects
            structure_chains = [x.id for x in structure.first_model.child_list]
            self.add_chain_ids(structure_chains)
            self.get_structure_seqs(structure.first_model)

            # Also add all chains to self.mapped_chains ONLY if there are none specified
            if not self.mapped_chains:
                self.add_mapped_chain_ids(structure_chains)

            return structure

    def clean_structure(self, out_suffix='_clean', outdir=None, force_rerun=False,
                        remove_atom_alt=True, remove_atom_hydrogen=True, keep_atom_alt_id='A', add_atom_occ=True,
                        remove_res_hetero=True, add_chain_id_if_empty='X', keep_chains=None):
        """Clean the structure file associated with this structure, and save it as a new file. Returns the file path.

        Args:
            out_suffix: Suffix to append to original filename
            outdir: Path to output directory
            force_rerun: If structure should be re-cleaned if a clean file exists already
            remove_atom_alt: Remove alternate positions
            remove_atom_hydrogen: Remove hydrogen atoms
            keep_atom_alt_id: If removing alternate positions, which alternate ID to keep
            add_atom_occ: Add atom occupancy fields if not present
            remove_res_hetero: Remove all HETATMs
            add_chain_id_if_empty: Add a chain ID if not present
            keep_chains: Keep only these chains

        Returns:
            str: Path to cleaned PDB file

        """

        if not self.structure_path:
            log.error('{}: no structure file, unable to clean'.format(self.id))
            return None

        clean_pdb_file = ssbio.protein.structure.utils.cleanpdb.clean_pdb(self.structure_path, out_suffix=out_suffix,
                                                                          outdir=outdir, force_rerun=force_rerun,
                                                                          remove_atom_alt=remove_atom_alt,
                                                                          remove_atom_hydrogen=remove_atom_hydrogen,
                                                                          keep_atom_alt_id=keep_atom_alt_id,
                                                                          add_atom_occ=add_atom_occ,
                                                                          remove_res_hetero=remove_res_hetero,
                                                                          add_chain_id_if_empty=add_chain_id_if_empty,
                                                                          keep_chains=keep_chains)

        return clean_pdb_file

    def add_mapped_chain_ids(self, mapped_chains):
        """Add chains by ID into the mapped_chains attribute

        Args:
            mapped_chains (str, list): Chain ID or list of IDs

        """
        mapped_chains = ssbio.utils.force_list(mapped_chains)

        for c in mapped_chains:
            if c not in self.mapped_chains:
                self.mapped_chains.append(c)
                log.debug('{}: added to list of mapped chains'.format(c))
            else:
                log.debug('{}: chain already in list of mapped chains, not adding'.format(c))

    def add_chain_ids(self, chains):
        """Add chains by ID into the chains attribute

        Args:
            chains (str, list): Chain ID or list of IDs

        """
        chains = ssbio.utils.force_list(chains)

        for c in chains:
            if self.chains.has_id(c):
                log.debug('{}: chain already present'.format(c))
            else:
                chain_prop = ChainProp(ident=c, pdb_parent=self.id)
                self.chains.append(chain_prop)
                log.debug('{}: added to chains list'.format(c))

    def get_structure_seqs(self, model):
        """Store chain sequences in the corresponding ChainProp objects in the chains attribute."""
        # Don't overwrite existing ChainProp objects
        dont_overwrite = []
        chains = list(model.get_chains())
        for x in chains:
            if self.chains.has_id(x.id):
                if self.chains.get_by_id(x.id).seq_record:
                    dont_overwrite.append(x.id)
        if len(dont_overwrite) == len(chains):
            log.debug('Not writing structure sequences, already stored')
            return

        # Returns the structures sequences with Xs added
        structure_seqs = ssbio.protein.structure.properties.residues.get_structure_seqrecords(model)
        log.debug('{}: gathered chain sequences'.format(self.id))

        # Associate with ChainProps
        for seq_record in structure_seqs:
            log.debug('{}: adding chain sequence to ChainProp'.format(seq_record.id))
            my_chain = self.chains.get_by_id(seq_record.id)
            my_chain.seq_record = seq_record

    def reset_chain_seq_records(self):
        for x in self.chains:
            x.reset_seq_record()

    def align_seqprop_to_mapped_chains(self, seqprop, outdir=None, engine='needle', parse=True, force_rerun=False,
                                       **kwargs):
        """Run and store alignments of the reference sequence to chains in mapped_chains.

        Alignments are stored in the seqprop.structure_alignments attribute.

        Args:
            outdir (str): Directory to output sequence alignment files (only if running with needle)
            engine (str): Which pairwise alignment tool to use ("needle" or "biopython")
            parse (bool): Store locations of mutations, insertions, and deletions in the alignment object (as an annotation)
            force_rerun:
            **kwargs: Other alignment options
        """
        # TODO: **kwargs for alignment options

        # Parse the structure so chain sequences are stored
        my_structure = self.parse_structure()

        for chain_id in self.mapped_chains:
            structure_id = '{}-{}'.format(self.id, chain_id)
            aln_id = '{}_{}'.format(seqprop.id, structure_id)
            outfile = '{}.needle'.format(aln_id)

            if seqprop.structure_alignments.has_id(aln_id):
                log.debug('{}: alignment already completed'.format(chain_id))
                continue

            log.debug('{}: aligning to reference sequence {}'.format(structure_id, seqprop.id))

            chain_prop = self.chains.get_by_id(chain_id)
            chain_seq_record = chain_prop.seq_record
            if not chain_seq_record:
                raise ValueError('{}: chain sequence not parsed'.format(chain_id))

            aln = ssbio.protein.sequence.utils.alignment.pairwise_sequence_alignment(a_seq=seqprop.seq_str,
                                                                                     a_seq_id=seqprop.id,
                                                                                     b_seq=chain_seq_record,
                                                                                     b_seq_id=structure_id,
                                                                                     engine=engine,
                                                                                     outdir=outdir,
                                                                                     outfile=outfile,
                                                                                     force_rerun=force_rerun)

            # Add an identifier to the MultipleSeqAlignment object for storage in a DictList
            aln.id = aln_id
            aln.annotations['a_seq'] = seqprop.id
            aln.annotations['b_seq'] = structure_id
            aln.annotations['structure_id'] = self.id
            aln.annotations['chain_id'] = chain_id

            if parse:
                aln_df = ssbio.protein.sequence.utils.alignment.get_alignment_df(a_aln_seq=str(list(aln)[0].seq),
                                                                                 b_aln_seq=str(list(aln)[1].seq))
                aln.annotations['mutations'] = ssbio.protein.sequence.utils.alignment.get_mutations(aln_df)
                aln.annotations['deletions'] = ssbio.protein.sequence.utils.alignment.get_deletions(aln_df)
                aln.annotations['insertions'] = ssbio.protein.sequence.utils.alignment.get_insertions(aln_df)

            seqprop.structure_alignments.append(aln)

    def sequence_quality_checker(self, reference_seq, seq_ident_cutoff=0.5, allow_missing_on_termini=0.2,
                                 allow_mutants=True, allow_deletions=False,
                                 allow_insertions=False, allow_unresolved=True):
        """Set the representative chain based on sequence quality checks to the reference sequence.

        Args:
            seq_ident_cutoff (float): Percent sequence identity cutoff, in decimal form
            allow_missing_on_termini (float): Percentage of the total length of the reference sequence which will be ignored
                when checking for modifications. Example: if 0.1, and reference sequence is 100 AA, then only residues
                5 to 95 will be checked for modifications.
            allow_mutants (bool): If mutations should be allowed or checked for
            allow_deletions (bool): If deletions should be allowed or checked for
            allow_insertions (bool): If insertions should be allowed or checked for
            allow_unresolved (bool): If unresolved residues should be allowed or checked for

        Returns:
            ChainProp: the best chain, if any

        """
        for chain_id in self.mapped_chains:
            try:
                alignment = reference_seq.structure_alignments.get_by_id('{}_{}-{}'.format(reference_seq.id, self.id, chain_id))
            except KeyError:
                log.error('{}_{}-{}: structure alignment not found'.format(reference_seq.id, self.id, chain_id))
                continue

            # Compare representative sequence to structure sequence using the alignment
            found_good_chain = ssbio.protein.structure.properties.quality.sequence_checker(reference_seq_aln=alignment[0],
                                                                                           structure_seq_aln=alignment[1],
                                                                                           seq_ident_cutoff=seq_ident_cutoff,
                                                                                           allow_missing_on_termini=allow_missing_on_termini,
                                                                                           allow_mutants=allow_mutants,
                                                                                           allow_deletions=allow_deletions,
                                                                                           allow_insertions=allow_insertions,
                                                                                           allow_unresolved=allow_unresolved)

            # If found_good_pdb = True, set as representative chain
            # If not, move on to the next potential chain
            if found_good_chain:
                self.representative_chain = self.chains.get_by_id(chain_id)
                self.reference_seq_top_coverage = alignment.annotations['percent_identity']
                log.debug('{}: chain {} set as representative'.format(self.id, chain_id))
                return self.representative_chain
        else:
            log.debug('{}: no chains meet quality checks'.format(self.id))
            return None

    def get_dict_with_chain(self, chain, only_keys=None, chain_keys=None, exclude_attributes=None, df_format=False):
        """get_dict method which incorporates attributes found in a specific chain. Does not overwrite any attributes
            in the original StructProp.

        Args:
            chain:
            only_keys:
            chain_keys:
            exclude_attributes:
            df_format:

        Returns:
            dict: attributes of StructProp + the chain specified

        """

        # Choose attributes to return, return everything in the object if a list is not specified
        if not only_keys:
            keys = list(self.__dict__.keys())
        else:
            keys = ssbio.utils.force_list(only_keys)

        # Remove keys you don't want returned
        if exclude_attributes:
            exclude_attributes = ssbio.utils.force_list(exclude_attributes)
            for x in exclude_attributes:
                if x in keys:
                    keys.remove(x)
        else:
            exclude_attributes = []

        exclude_attributes.extend(['mapped_chains', 'chains'])

        final_dict = {k: v for k, v in Object.get_dict(self, only_keys=keys, exclude_attributes=exclude_attributes,
                                                       df_format=df_format).items()}


        chain_prop = self.chains.get_by_id(chain)
        # Filter out keys that show up in StructProp
        if not chain_keys:
            chain_keys = [x for x in chain_prop.get_dict().keys() if x not in final_dict]

        chain_dict = chain_prop.get_dict(only_keys=chain_keys, df_format=df_format)
        final_dict.update(chain_dict)

        return final_dict

    def get_disulfide_bridges(self, threshold=3.0):
        """Run Biopython's search_ss_bonds to find potential disulfide bridges for each chain and store in ChainProp.
        """
        parsed = self.parse_structure()
        if not parsed:
            log.error('{}: unable to open structure to find S-S bridges'.format(self.id))
            return

        disulfide_bridges = ssbio.protein.structure.properties.residues.search_ss_bonds(parsed.first_model,
                                                                                        threshold=threshold)
        if not disulfide_bridges:
            log.debug('{}: no disulfide bridges'.format(self.id))

        for chain, bridges in disulfide_bridges.items():
            self.representative_chain.seq_record.annotations['SSBOND-biopython'] = disulfide_bridges[self.representative_chain.id]
            log.debug('{}: found {} disulfide bridges'.format(chain, len(bridges)))
            log.debug('{}: stored disulfide bridges in seq_record letter_annotations'.format(chain))

    def get_residue_depths(self, outdir, force_rerun=False):
        """Run MSMS on this structure and store the residue depths/ca depths in the corresponding ChainProp SeqRecords
        """
        parsed = self.parse_structure()
        if not parsed:
            log.error('{}: unable to open structure to run MSMS'.format(self.id))
            return

        log.debug('{}: running MSMS'.format(self.id))
        msms_results = ssbio.protein.structure.properties.msms.get_msms_df(model=parsed.first_model,
                                                                           pdb_file=self.structure_path,
                                                                           outdir=outdir, force_rerun=force_rerun)
        if msms_results.empty:
            log.error('{}: unable to run MSMS'.format(self.id))
            return

        chains = msms_results.chain.unique()

        for chain in chains:
            res_depths = msms_results[msms_results.chain == chain].res_depth.tolist()
            ca_depths = msms_results[msms_results.chain == chain].ca_depth.tolist()

            chain_prop = self.chains.get_by_id(chain)
            chain_seq = chain_prop.seq_record

            # Making sure the X's are filled in
            res_depths = ssbio.protein.structure.properties.residues.match_structure_sequence(orig_seq=chain_seq,
                                                                                              new_seq=res_depths,
                                                                                              fill_with=float('Inf'))

            ca_depths = ssbio.protein.structure.properties.residues.match_structure_sequence(orig_seq=chain_seq,
                                                                                             new_seq=ca_depths,
                                                                                             fill_with=float('Inf'))

            chain_prop.seq_record.letter_annotations['RES_DEPTH-msms'] = res_depths
            chain_prop.seq_record.letter_annotations['CA_DEPTH-msms'] = ca_depths
            log.debug('{}: stored residue depths in chain seq_record letter_annotations'.format(chain))

    def get_dssp_annotations(self, outdir, force_rerun=False):
        """Run DSSP on this structure and store the DSSP annotations in the corresponding ChainProp SeqRecords

        Args:
            outdir (str): Path to where DSSP dataframe will be stored.
            force_rerun (bool): If DSSP results should be recalculated

        """
        parsed = self.parse_structure()
        if not parsed:
            log.error('{}: unable to open structure to run DSSP'.format(self.id))
            return

        log.debug('{}: running DSSP'.format(self.id))
        dssp_results = ssbio.protein.structure.properties.dssp.get_dssp_df(model=parsed.first_model,
                                                                           pdb_file=self.structure_path,
                                                                           outdir=outdir,
                                                                           force_rerun=force_rerun)

        if dssp_results.empty:
            log.error('{}: unable to run DSSP'.format(self.id))
            return

        chains = dssp_results.chain.unique()
        dssp_summary = ssbio.protein.structure.properties.dssp.secondary_structure_summary(dssp_results)

        for chain in chains:
            ss = dssp_results[dssp_results.chain == chain].ss.tolist()
            exposure_rsa = dssp_results[dssp_results.chain == chain].exposure_rsa.tolist()
            exposure_asa = dssp_results[dssp_results.chain == chain].exposure_asa.tolist()
            phi = dssp_results[dssp_results.chain == chain].phi.tolist()
            psi = dssp_results[dssp_results.chain == chain].psi.tolist()

            chain_prop = self.chains.get_by_id(chain)
            chain_seq = chain_prop.seq_record

            # Making sure the X's are filled in
            ss = ssbio.protein.structure.properties.residues.match_structure_sequence(orig_seq=chain_seq,
                                                                                      new_seq=ss,
                                                                                      fill_with='-')
            exposure_rsa = ssbio.protein.structure.properties.residues.match_structure_sequence(orig_seq=chain_seq,
                                                                                                new_seq=exposure_rsa,
                                                                                                fill_with=float('Inf'))
            exposure_asa = ssbio.protein.structure.properties.residues.match_structure_sequence(orig_seq=chain_seq,
                                                                                                new_seq=exposure_asa,
                                                                                                fill_with=float('Inf'))
            phi = ssbio.protein.structure.properties.residues.match_structure_sequence(orig_seq=chain_seq,
                                                                                       new_seq=phi,
                                                                                       fill_with=float('Inf'))
            psi = ssbio.protein.structure.properties.residues.match_structure_sequence(orig_seq=chain_seq,
                                                                                       new_seq=psi,
                                                                                       fill_with=float('Inf'))

            chain_prop.seq_record.annotations.update(dssp_summary[chain])

            chain_prop.seq_record.letter_annotations['SS-dssp'] = ss
            chain_prop.seq_record.letter_annotations['RSA-dssp'] = exposure_rsa
            chain_prop.seq_record.letter_annotations['ASA-dssp'] = exposure_asa
            chain_prop.seq_record.letter_annotations['PHI-dssp'] = phi
            chain_prop.seq_record.letter_annotations['PSI-dssp'] = psi
            log.debug('{}: stored DSSP annotations in chain seq_record letter_annotations'.format(chain))

    def _map_repseq_resnums_to_repchain_index(self, reference_seq, resnums):
        """Map a residue number in the reference_seq to an index in the representative_chain

        Use this to get the indices of the repchain to get structural properties at a specific residue number.

        Args:
            resnums (int, list): Residue numbers in the representative sequence

        Returns:
            dict: Mapping of resnums to indices

        """
        resnums = ssbio.utils.force_list(resnums)

        repchain_resnum_mapping = reference_seq.seq_record.letter_annotations['repchain_resnums']

        to_repchain_index = {}
        for x in resnums:
            ix = repchain_resnum_mapping[x - 1] - 1

            if np.isnan(ix):
                log.warning('{}, {}: no equivalent residue found in structure sequence'.format(self.id, x))
            else:
                to_repchain_index[x] = int(ix)

        return to_repchain_index

    def map_repseq_resnums_to_structure_resnums(self, reference_seq, resnums, chain=None):
        """Map a residue number in the reference_seq to the actual structure file's residue number (for the representative chain

        Args:
            resnums (int, list): Residue numbers in the representative sequence

        Returns:
            dict: Mapping of resnums to structure residue IDs

        """
        if chain:
            chain = self.chains.get_by_id(chain)
        else:
            chain = self.representative_chain

        resnums = ssbio.utils.force_list(resnums)

        mapping_to_repchain_index = self._map_repseq_resnums_to_repchain_index(reference_seq, resnums)
        repchain_structure_mapping = chain.seq_record.letter_annotations['structure_resnums']

        to_structure_resnums = {}
        for k, v in mapping_to_repchain_index.items():
            rn = repchain_structure_mapping[v]

            if rn[1] == float('Inf'):
                log.warning('{}, {}: structure file does not contain coordinates for this residue'.format(self.id, k))
            else:
                to_structure_resnums[k] = rn

        return to_structure_resnums

    def view_structure(self, opacity=1.0, recolor=True, gui=False):
        """Use NGLviewer to display a structure in a Jupyter notebook

        Args:
            opacity (float): Opacity of the structure
            gui (bool): If the NGLview GUI should show up

        Returns:
            NGLviewer object

        """
        # TODO: test other ways we can manipulate the view object

        if not self.structure_path:
            raise ValueError("Structure file not loaded")
        view = nv.show_structure_file(self.structure_path, gui=gui)

        if recolor:
            view.clear_representations()
            view.add_cartoon(selection='protein', color='silver', opacity=opacity)
        # else:
            # view.add_cartoon(selection='protein', opacity=opacity)
        return view

    def view_structure_and_highlight_residues(self, structure_resnums, chain=None, color='red',
                                              structure_opacity=0.5, gui=False):
        """Input a residue number or numbers to view on the structure.

        Args:
            structure_resnums (int, list): Residue number(s) to highlight
            chain (str, list): Chain ID or IDs of which residues are a part of. If not provided, all chains in the
                mapped_chains attribute will be used. PLEASE NOTE: if that is also empty, all residues in all chains
                matching the residue numbers will be shown.
            color (str): Color to highlight with
            structure_opacity (float): Opacity of the protein structure cartoon representation
            gui (bool): If the NGLview GUI should show up

        Returns:
            NGLviewer object

        """
        if not chain:
            chain = self.mapped_chains
            if not chain:
                parsed = self.parse_structure()
                if not parsed:
                    log.error('{}: unable to open structure to get chains'.format(self.id))
                    return
                log.warning('Showing all residue numbers on all chains, please input chain if desired')

        view = self.view_structure(opacity=structure_opacity, gui=gui)

        if isinstance(structure_resnums, list):
            unique_mutations = list(set(structure_resnums))
        elif isinstance(structure_resnums, int):
            unique_mutations = ssbio.utils.force_list(structure_resnums)

        # TODO: add color by letter_annotations!

        colors = sns.color_palette("hls", len(unique_mutations)).as_hex()

        to_show_chains = '( '
        for c in chain:
            to_show_chains += ':{} or'.format(c)
        to_show_chains = to_show_chains.strip(' or ')
        to_show_chains += ' )'

        to_show_res = '( '
        for m in unique_mutations:
            to_show_res += '{} or '.format(m)
        to_show_res = to_show_res.strip(' or ')
        to_show_res += ' )'

        log.info('Selection: {} and not hydrogen and {}'.format(to_show_chains, to_show_res))

        view.add_ball_and_stick(selection='{} and not hydrogen and {}'.format(to_show_chains, to_show_res), color=color)

        return view

    def view_structure_and_highlight_residues_scaled(self, structure_resnums, chain=None, color='red', unique_colors=False,
                                                     structure_opacity=0.5, opacity_range=(0.5,1), scale_range=(.7, 10),
                                                     gui=False):
        """Input a list of residue numbers to view on the structure. Or input a dictionary of residue numbers to counts
            to scale residues by counts (useful to view mutations).

        Args:
            structure_resnums (int, list, dict): Residue number(s) to highlight, or
                a dictionary of residue number to frequency count
            chain (str, list): Chain ID or IDs of which residues are a part of. If not provided, all chains in the
                mapped_chains attribute will be used. PLEASE NOTE: if that is also empty, all residues in all chains
                matching the residue numbers will be shown.
            color (str): Color to highlight with
            unique_colors (bool): If each mutation should be colored uniquely (will override color argument)
            structure_opacity (float): Opacity of the protein structure cartoon representation
            opacity_range (tuple): Min/max opacity values (residues that have higher frequency counts will be opaque)
            scale_range (tuple): Min/max size values (residues that have higher frequency counts will be bigger)
            gui (bool): If the NGLview GUI should show up

        Returns:
            NGLviewer object

        """
        if not chain:
            chain = self.mapped_chains
            if not chain:
                parsed = self.parse_structure()
                if not parsed:
                    log.error('{}: unable to open structure to get chains'.format(self.id))
                    return
                log.warning('Showing all residue numbers on all chains, please input chain if desired')
        else:
            chain = ssbio.utils.force_list(chain)

        if isinstance(structure_resnums, dict):
            opacity_dict = ssbio.utils.scale_calculator(opacity_range[0], structure_resnums, rescale=opacity_range)
            scale_dict = ssbio.utils.scale_calculator(scale_range[0], structure_resnums, rescale=scale_range)
        else:
            opacity_dict = {x: max(opacity_range) for x in ssbio.utils.force_list(structure_resnums)}
            scale_dict = {x: max(scale_range) for x in ssbio.utils.force_list(structure_resnums)}

        view = self.view_structure(opacity=structure_opacity, gui=gui)

        if isinstance(structure_resnums, list):
            unique_mutations = list(set(structure_resnums))
        elif isinstance(structure_resnums, dict):
            unique_mutations = list(structure_resnums.keys())
        elif isinstance(structure_resnums, int):
            unique_mutations = ssbio.utils.force_list(structure_resnums)

        # TODO: add color by letter_annotations!

        colors = sns.color_palette("hls", len(unique_mutations)).as_hex()

        to_show_chains = '( '
        for c in chain:
            to_show_chains += ':{} or'.format(c)
        to_show_chains = to_show_chains.strip(' or ')
        to_show_chains += ' )'

        for i, x in enumerate(unique_mutations):
            if isinstance(x, tuple):
                to_show_res = '( '
                for mut in x:
                    to_show_res += '{} or '.format(mut)
                to_show_res = to_show_res.strip(' or ')
                to_show_res += ' )'
            else:
                to_show_res = x

            log.info('Selection: {} and not hydrogen and {}'.format(to_show_chains, to_show_res))

            if unique_colors:
                view.add_ball_and_stick(selection='{} and not hydrogen and {}'.format(to_show_chains, to_show_res),
                                        color=colors[i], opacity=opacity_dict[x], scale=scale_dict[x])
            else:
                view.add_ball_and_stick(selection='{} and not hydrogen and {}'.format(to_show_chains, to_show_res),
                                        color=color, opacity=opacity_dict[x], scale=scale_dict[x])

        return view

    def __json_decode__(self, **attrs):
        for k, v in attrs.items():
            if k == 'chains':
                setattr(self, k, DictList(v))
            else:
                setattr(self, k, v)


from json_tricks.np import dump, load
import pickle
import logging
log = logging.getLogger(__name__)


def save_json(obj, outfile, allow_nan=True, compression=True):
    """Save an ssbio object as a JSON file using json_tricks"""
    if compression:
        with open(outfile, 'wb') as f:
            dump(obj, f, allow_nan=allow_nan, compression=compression)
    else:
        with open(outfile, 'w') as f:
            dump(obj, f, allow_nan=allow_nan, compression=compression)
    log.info('Saved {} (id: {}) to {}'.format(type(obj), obj.id, outfile))


def load_json(file, new_root_dir=None, decompression=True):
    """Load a JSON file using json_tricks"""
    if decompression:
        with open(file, 'rb') as f:
            my_object = load(f, decompression=decompression)
    else:
        with open(file, 'r') as f:
            my_object = load(f, decompression=decompression)
    if new_root_dir:
        my_object.root_dir = new_root_dir

    return my_object


def save_pickle(obj, outfile, protocol=2):
    """Save the object as a pickle file

    Args:
        outfile (str): Filename
        protocol (int): Pickle protocol to use. Default is 2 to remain compatible with Python 2

    Returns:
        str: Path to pickle file

    """
    with open(outfile, 'wb') as f:
        pickle.dump(obj, f, protocol=protocol)

    return outfile


def load_pickle(file):
    """Load a pickle file.
    
    Args:
        file (str): Path to pickle file

    Returns:
        object: Loaded object from pickle file

    """
    with open(file, 'rb') as f:
       return pickle.load(f)


import logging
import os.path as op

import ssbio.protein.sequence.utils.fasta
import ssbio.utils

log = logging.getLogger(__name__)


class SCRATCH():
    """Provide wrappers for running and parsing SCRATCH on a sequence file or sequence string.

    SCRATCH predicts:
        - Secondary structure
            - 3 classes (helix, strand, other) using SSpro
            - 8 classes (standard DSSP definitions) using SSpro8
        - Relative solvent accessibility (RSA, also known as relative accessible surface area)
            - @ 25% exposed RSA cutoff (<25% RSA means it is buried)
            - @ all cutoffs in 5% increments from 0 to 100
    """

    # TODO: also provide summary dataframes

    def __init__(self, project_name, seq_file=None, seq_str=None):
        self.project_name = project_name
        self.seq_file = seq_file
        if seq_str:
            self.seq_file = ssbio.protein.sequence.utils.fasta.write_seq_as_temp_fasta(seq_str)

    def run_scratch(self, path_to_scratch, num_cores=1, outname=None, outdir=None, force_rerun=False):
        """Run SCRATCH on the sequence_file that was loaded into the class.

        Args:
            path_to_scratch: Path to the SCRATCH executable, run_SCRATCH-1D_predictors.sh
            outname: Prefix to name the output files
            outdir: Directory to store the output files
            force_rerun: Flag to force rerunning of SCRATCH even if the output files exist

        Returns:

        """
        if not outname:
            outname = self.project_name
        if not outdir:
            outdir = ''

        outname = op.join(outdir, outname)

        self.out_sspro = '{}.ss'.format(outname)
        self.out_sspro8 = '{}.ss8'.format(outname)
        self.out_accpro = '{}.acc'.format(outname)
        self.out_accpro20 = '{}.acc20'.format(outname)

        # TODO: check for multiple output files in command_runner
        ssbio.utils.command_runner(shell_command='{} {} {} {}'.format(path_to_scratch, self.seq_file, outname, num_cores),
                                   outfile='{}.ss'.format(outname),
                                   force_rerun_flag=force_rerun)

    def sspro_results(self):
        """Parse the SSpro output file and return a dict of secondary structure compositions.

        Returns:
            dict: Keys are sequence IDs, values are the lists of secondary structure predictions.
                H: helix
                E: strand
                C: the rest

        """
        return ssbio.protein.sequence.utils.fasta.load_fasta_file_as_dict_of_seqs(self.out_sspro)

    def sspro_summary(self):
        """Parse the SSpro output file and return a summary of secondary structure composition.

        The output file is just a FASTA formatted file, so you can get residue level
            information by parsing it like a normal sequence file.

        Returns:
            dict: Percentage of:
                H: helix
                E: strand
                C: the rest

        """
        summary = {}

        records = ssbio.protein.sequence.utils.fasta.load_fasta_file(self.out_sspro)
        for r in records:
            seq_summary = {}
            seq_summary['percent_H-sspro'] = r.seq.count('H')/float(len(r))
            seq_summary['percent_E-sspro'] = r.seq.count('E')/float(len(r))
            seq_summary['percent_C-sspro'] = r.seq.count('C')/float(len(r))

            summary[r.id] = seq_summary

        return summary

    def sspro8_results(self):
        """Parse the SSpro8 output file and return a dict of secondary structure compositions.
        """
        return ssbio.protein.sequence.utils.fasta.load_fasta_file_as_dict_of_seqs(self.out_sspro8)

    def sspro8_summary(self):
        """Parse the SSpro8 output file and return a summary of secondary structure composition.

        The output file is just a FASTA formatted file, so you can get residue level
            information by parsing it like a normal sequence file.

        Returns:
            dict: Percentage of:
                H: alpha-helix
                G: 310-helix
                I: pi-helix (extremely rare)
                E: extended strand
                B: beta-bridge
                T: turn
                S: bend
                C: the rest

        """
        summary = {}

        records = ssbio.protein.sequence.utils.fasta.load_fasta_file(self.out_sspro8)
        for r in records:
            seq_summary = {}
            seq_summary['percent_H-sspro8'] = r.seq.count('H') / float(len(r))
            seq_summary['percent_G-sspro8'] = r.seq.count('G') / float(len(r))
            seq_summary['percent_I-sspro8'] = r.seq.count('I') / float(len(r))
            seq_summary['percent_E-sspro8'] = r.seq.count('E') / float(len(r))
            seq_summary['percent_B-sspro8'] = r.seq.count('B') / float(len(r))
            seq_summary['percent_T-sspro8'] = r.seq.count('T') / float(len(r))
            seq_summary['percent_S-sspro8'] = r.seq.count('S') / float(len(r))
            seq_summary['percent_C-sspro8'] = r.seq.count('C') / float(len(r))

            summary[r.id] = seq_summary

        return summary

    def accpro_results(self):
        """Parse the ACCpro output file and return a dict of secondary structure compositions.
        """
        return ssbio.protein.sequence.utils.fasta.load_fasta_file_as_dict_of_seqs(self.out_accpro)

    def accpro_summary(self):
        """Parse the ACCpro output file and return a summary of percent exposed/buried residues.

        The output file is just a FASTA formatted file, so you can get residue level
            information by parsing it like a normal sequence file.

        Returns:
            dict: Percentage of buried and exposed residues

        """
        summary = {}

        records = ssbio.protein.sequence.utils.fasta.load_fasta_file(self.out_accpro)
        for r in records:
            seq_summary = {}
            seq_summary['percent_exposed-accpro'] = r.seq.count('e') / float(len(r))
            seq_summary['percent_buried-accpro'] = r.seq.count('-') / float(len(r))

            summary[r.id] = seq_summary

        return summary

    def accpro20_results(self):
        """Parse the ACCpro output file and return a dict of secondary structure compositions"""
        return read_accpro20(self.out_accpro20)
        # return ssbio.sequence.utils.fasta.load_fasta_file_as_dict_of_seqs(self.out_accpro20)

    def accpro20_summary(self, cutoff):
        """Parse the ACCpro output file and return a summary of percent exposed/buried residues based on a cutoff.

        Below the cutoff = buried
        Equal to or greater than cutoff = exposed
        The default cutoff used in accpro is 25%.

        The output file is just a FASTA formatted file, so you can get residue level
            information by parsing it like a normal sequence file.

        Args:
            cutoff (float): Cutoff for defining a buried or exposed residue.

        Returns:
            dict: Percentage of buried and exposed residues

        """
        summary = {}

        if cutoff < 1:
            cutoff = 1 * 100

        records = read_accpro20(self.out_accpro20)

        for k,v in records.items():
            seq_summary = {}

            exposed = 0
            buried = 0
            for s in v:
                if s > cutoff:
                    exposed += 1
                else:
                    buried += 1

            seq_summary['percent_exposed-accpro20'] = exposed / float(len(v))
            seq_summary['percent_buried-accpro20'] = buried / float(len(v))

            summary[k] = seq_summary

        return summary


def read_accpro20(infile):
    """Read the accpro20 output (.acc20) and return the parsed FASTA records.

    Keeps the spaces between the accessibility numbers.

    Args:
        infile: Path to .acc20 file

    Returns:
        dict: Dictionary of accessibilities with keys as the ID

    """
    with open(infile) as f:
        records = f.read().splitlines()

    accpro20_dict = {}
    for i, r in enumerate(records):
        if i % 2 == 0:
            # TODO: Double check how to parse FASTA IDs (can they have a space because that is what i split by)
            # Key was originally records[i][1:]
            accpro20_dict[records[i].split(' ')[0][1:]] = [int(x) for x in records[i + 1].split(' ')]

    return accpro20_dict


import unittest

import ssbio.databases.isoform


class TestIsoform(unittest.TestCase):
    """Unit tests for PDB
    """

    def test_gene_isos_to_uniprot(self):
        gene_1 = '52'
        isos_1 = ['1','2','3']
        uniprot_isos_1 = ['P24666-1','P24666-2','P24666-3']
        unrev_uniprot_isos_1 = ['A0A0B4J207-1']
        expected_1 = {'1':'P24666-1','2':'P24666-2','3':'P24666-3'}
        self.assertEqual(ssbio.databases.isoform.gene_isos_to_uniprot(gene_1, isos_1, uniprot_isos_1, unrev_uniprot_isos_1), expected_1)

        gene_2 = '52'
        isos_2 = ['1', '2', '3']
        uniprot_isos_2 = ['P24666-1', 'P24666-2']
        unrev_uniprot_isos_2 = []
        expected_2 = {'1': 'P24666-1', '2': 'P24666-2', '3': None}
        self.assertEqual(ssbio.databases.isoform.gene_isos_to_uniprot(gene_2, isos_2, uniprot_isos_2, unrev_uniprot_isos_2), expected_2)

        gene_3 = '221823'
        isos_3 = ['1', '2']
        uniprot_isos_3 = ['P21108-1']
        unrev_uniprot_isos_3 = ['A0A0B4J207-1']
        expected_3 = {'1': 'P21108-1', '2': 'A0A0B4J207-1'}
        self.assertEqual(ssbio.databases.isoform.gene_isos_to_uniprot(gene_3, isos_3, uniprot_isos_3, unrev_uniprot_isos_3), expected_3)


import unittest
import six
import ssbio.databases.bigg

class TestBiGG(unittest.TestCase):
    """Unit tests for BiGG
    """

    def test_get_pdbs_for_gene(self):
        model = 'e_coli_core'
        gene = 'b0118'

        expected = [('1l5j', 'A'), ('1l5j', 'B')]

        six.assertCountEqual(self, expected, ssbio.databases.bigg.get_pdbs_for_gene(model, gene))

        model = 'e_coli_core'
        gene = 'b0351'

        expected = []

        six.assertCountEqual(self, expected, ssbio.databases.bigg.get_pdbs_for_gene(model, gene))


import os.path as op
import unittest

from Bio.Align import MultipleSeqAlignment

import ssbio.protein.sequence.utils.alignment


class TestAlignment(unittest.TestCase):
    """Unit tests for Alignment"""

    def test_pairwise_sequence_alignment_str(self):
        base_id = 'a_test'
        base = 'MSPRVGVTLSGRYRLQRLIATGGMGQVWEAVDNRLGRRVAVVASASA'
        muts_id = 'b_test'
        muts = 'MSPRVGVTLSGRYRLQRLIATGGMGQVWEAVDNRLGRRVARRVAASSSAS'

        alignment = ssbio.protein.sequence.utils.alignment.pairwise_sequence_alignment(a_seq_id=base_id, a_seq=base,
                                                                                       b_seq_id=muts_id, b_seq=muts,
                                                                                       engine='biopython')
        self.assertTrue(isinstance(alignment, MultipleSeqAlignment))
        self.assertEqual(alignment[0].id, base_id)
        self.assertEqual(alignment[0].seq, 'MSPRVGVTLSGRYRLQRLIATGGMGQVWEAVDNRLGRRVAV--VA-SASA-')
        self.assertEqual(alignment[1].id, muts_id)
        self.assertEqual(alignment[1].seq, 'MSPRVGVTLSGRYRLQRLIATGGMGQVWEAVDNRLGRRVA-RRVAASSSAS')
        self.assertEqual(alignment.annotations, {'end': 51, 'percent_identity': 88.23529411764706, 'score': 224.0, 'start': 0})

        alignment2 = ssbio.protein.sequence.utils.alignment.pairwise_sequence_alignment(a_seq_id=base_id, a_seq=base,
                                                                                        b_seq_id=muts_id, b_seq=muts,
                                                                                        engine='needle', force_rerun=True)
        self.assertTrue(isinstance(alignment2, MultipleSeqAlignment))
        self.assertEqual(alignment2[0].id, base_id)
        self.assertEqual(alignment2[0].seq, 'MSPRVGVTLSGRYRLQRLIATGGMGQVWEAVDNRLGRRVA--VVASASA-')
        self.assertEqual(alignment2[1].id, muts_id)
        self.assertEqual(alignment2[1].seq, 'MSPRVGVTLSGRYRLQRLIATGGMGQVWEAVDNRLGRRVARRVAASSSAS')
        self.assertEqual(alignment2.annotations, {'score': 213.5, 'percent_gaps': 6.0, 'percent_similarity': 92.0, 'percent_identity': 90.0})

    def test_run_needle_alignment_on_files(self):
        """Test if needle alignment runs correctly on 2 input files

        """
        outdir = op.join('test_files', 'out')
        working_dir = op.join('test_files', 'sequences')
        infile_a = op.join(working_dir, 'P9WGE7.fasta')
        infile_b = op.join(working_dir, '1gn3_A.faa')
        test_outfile = op.join(working_dir, 'P9WGE7_1gn3_A_align.txt')

        outfile = ssbio.protein.sequence.utils.alignment.run_needle_alignment_on_files(id_a='P9WGE7', faa_a=infile_a,
                                                                                       id_b='1gn3_A', faa_b=infile_b,
                                                                                       outdir=outdir,
                                                                                       outfile='test_alignment.txt',
                                                                                       force_rerun=True)

        # test if the file content is correct
        # ignore the first 12 lines because it is dated
        with open(outfile) as f:
            test_lines = f.readlines()[12:]
        with open(test_outfile) as f:
            correct_lines = f.readlines()[12:]

        self.assertEqual(test_lines, correct_lines)


    def test_run_needle_alignment_on_str(self):
        """Test if needle alignment runs correctly on 2 input strings

        """
        outdir = op.join('test_files', 'out')
        str1 = 'MAEYTLPDLDWDYGALEPHISGQINELHHSKHHATYVKGANDAVAKLEEA'
        str2 = 'XAEYTLPDLDWDYGALEPHISGQINELYHSKHHANDVKGANDAVAKLEEA'

        result = ['',
                  '#=======================================',
                  '#',
                  '# Aligned_sequences: 2',
                  '# 1: asis',
                  '# 2: asis',
                  '# Matrix: EBLOSUM62',
                  '# Gap_penalty: 10.0',
                  '# Extend_penalty: 0.5',
                  '#',
                  '# Length: 50',
                  '# Identity:      46/50 (92.0%)',
                  '# Similarity:    47/50 (94.0%)',
                  '# Gaps:           0/50 ( 0.0%)',
                  '# Score: 245.0',
                  '# ',
                  '#',
                  '#=======================================',
                  '',
                  'asis               1 MAEYTLPDLDWDYGALEPHISGQINELHHSKHHATYVKGANDAVAKLEEA     50',
                  '                     .||||||||||||||||||||||||||:||||||..||||||||||||||',
                  'asis               1 XAEYTLPDLDWDYGALEPHISGQINELYHSKHHANDVKGANDAVAKLEEA     50',
                  '',
                  '',
                  '#---------------------------------------',
                  '#---------------------------------------']

        outfile = op.join(outdir, 'test_alignment2.txt')

        outpath = ssbio.protein.sequence.utils.alignment.run_needle_alignment(seq_a=str1, seq_b=str2, outfile=outfile)

        # test if the result is the same
        # ignore the first 12 lines because it is dated
        with open(outpath, 'r') as f:
            raw_txt = f.read().splitlines()[12:]

        self.assertEqual(raw_txt, result)

        self.assertTrue(op.exists(outfile))


from ssbio import utils
date = utils.Date()

import os
import urllib.request
import io
import gzip
import os.path as op

matador_tsv_gz = "http://matador.embl.de/media/download/matador.tsv.gz"

def download_matador_db(outdir=None):
    if not outdir:
        outdir = os.getcwd()

    rename_unzipped_file = "{}-matador_full_database.tsv".format(date.short_date)

    if outdir:
        rename_unzipped_file = op.join(outdir, rename_unzipped_file)

    response = urllib.request.urlopen(matador_tsv_gz)
    compressed_file = io.BytesIO(response.read())
    decompressed_file = gzip.GzipFile(fileobj=compressed_file, mode='w')

    with open(rename_unzipped_file, "wt") as out_file:
        out_file.write(decompressed_file)

    return rename_unzipped_file